// SPDX-License-Identifier: MIT
/**
 * Initiate self refresh feature for Micron's eMMC device.
 * Copyright (c) 2021 Atmark Techno,Inc.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <stdint.h>
#include <unistd.h>
#include <stdbool.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <libgen.h>

#include <linux/mmc/ioctl.h>

/* regulated in JESD84-B51 */
#define EMMC_RSP_PRESENT (1 << 0)
#define EMMC_RSP_CRC (1 << 2)
#define EMMC_RSP_OPCODE (1 << 4)
#define EMMC_CMD_ADTC (1 << 5)
#define EMMC_RSP_SPI_S1 (1 << 7)
#define EMMC_RSP_SPI_R1 (EMMC_RSP_SPI_S1)
#define EMMC_RSP_R1 (EMMC_RSP_PRESENT | EMMC_RSP_CRC | EMMC_RSP_OPCODE)

#define EMMC_IOC_CMD _IOWR(MMC_BLOCK_MAJOR, 0, struct mmc_ioc_cmd)

#define EMMC_CMD_SET_TIME 49
#define EMMC_CMD_GEN_CMD 56
#define SSR_BUFF_SIZE 512 /** Secure Smart Report */
#define GEN_CMD_SSR 0x39 /** GEN_CMD arg for SSR feature */
#define OFFSET_CDATE 88 /** Offset bytes for completion date */
#define SECONDS_IN_A_DAY 86401
/** Offset seconds to convert to Gregorian time from epoch time */
#define EPOCH_OFFSET 62167219200
#define RTC_INFO_TYPE 1 /** payload block type for SET_TIME */

#define SUPPORTED_MID 0x13 /** micron */
#define SUPPORTED_REV 0x08 /** eMMC v5.1 */

static int use_verbose = 0;

/**
 * Send SET_TIME command.
 *  @param [in] dev       device file path. Set only eMMC 5.1 or later devices.
 *  @param [in] abs_time  seconds from Jan.1st, year 0 AD till current time in UTC.
 *  @return succeeded(0)/failed(!0)
 */
static int set_time(const char *dev, __u64 abs_time)
{
	struct mmc_ioc_cmd odata;
	int fd, ret = 0;
	struct {
		__u8 version;
		__u8 info_type;
		__u64 abs_time;
		__u8 reserved[502];
	} __attribute__((packed)) rtc_info_blk = {
		/** regulated in JESD84-B51 */
		.version = 1, /** version structure: shall be set to 1 */
		.reserved = { 0 }, /** shall be set to all zero */
	};

	fd = open(dev, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "Failed to open %s. %s\n", dev,
			strerror(errno));
		return 1;
	}

	/* pack mmc_ioc_cmd */
	memset(&odata, 0, sizeof(struct mmc_ioc_cmd));
	odata.write_flag = 1;
	odata.opcode = EMMC_CMD_SET_TIME;
	odata.arg = 0;
	odata.flags = EMMC_RSP_SPI_R1 | EMMC_RSP_R1 | EMMC_CMD_ADTC;
	odata.blksz = sizeof(rtc_info_blk);
	odata.blocks = 1;

	/* pack rtc_info_blk */
	rtc_info_blk.info_type = RTC_INFO_TYPE;
	rtc_info_blk.abs_time = abs_time;
	mmc_ioc_cmd_set_data(odata, &rtc_info_blk);

	ret = ioctl(fd, EMMC_IOC_CMD, &odata);
	close(fd);
	if (ret) {
		fprintf(stderr, "Failed to ioctl(CMD49) %s %s\n", dev,
			strerror(errno));
		return 2;
	}

	return 0;
}

/**
 * Get last self refresh completion date.
 *  @param [in]  dev       device file path. Set only eMMC 5.1 or later devices.
 *  @param [out] abs_time  seconds from Jan.1st, year 0 AD till current time in UTC.
 *  @return succeeded(0)/failed(!0)
 */
static int get_time(const char *dev, __u64 *abs_time)
{
	struct mmc_ioc_cmd idata;
	int fd, ret = 0;
	__u8 ssr_buf[SSR_BUFF_SIZE] = { 0 };

	fd = open(dev, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "Failed to open() %s. %s\n", dev,
			strerror(errno));
		return 1;
	}

	/* pack mmc_ioc_cmd */
	memset(&idata, 0, sizeof(struct mmc_ioc_cmd));
	idata.write_flag = 0;
	idata.opcode = EMMC_CMD_GEN_CMD;
	idata.arg = GEN_CMD_SSR;
	idata.flags = EMMC_RSP_SPI_R1 | EMMC_RSP_R1 | EMMC_CMD_ADTC;
	idata.blksz = SSR_BUFF_SIZE;
	idata.blocks = 1;
	mmc_ioc_cmd_set_data(idata, &ssr_buf);

	ret = ioctl(fd, EMMC_IOC_CMD, &idata);
	close(fd);
	if (ret) {
		fprintf(stderr, "Failed to ioctl(CMD56) %s %s\n", dev,
			strerror(errno));
		return 2;
	}

	*abs_time = *((__u64 *)&ssr_buf[OFFSET_CDATE]);

	if (use_verbose)
		printf("Last completion date: %llu\n", *abs_time);

	return 0;
}

/**
 * Read mmc block device attribute from sysfs.
 *  @param [in]  mmc_num  device number to read attribute.
 *  @param [in]  attr_name attribute name of mmc block device.
 *  @param [in,out]  attr  attribute string of mmc block device. result of this function.
 *  @param [in]  size      size of attr array. 
 *  @return succeeded(0)/failed(!0)
 */
static int read_attr(int mmc_num, const char *attr_name, char *attr,
		     size_t size)
{
	char path[PATH_MAX];
	int result;
	size_t len;
	char *res_fgets, *start = attr;
	FILE *fp;

	/**
     * note: support only when RCA is 0x0001 (linux initial value). 
     */
	result = snprintf(path, PATH_MAX,
			  "/sys/class/mmc_host/mmc%d/mmc%d:0001/%s", mmc_num,
			  mmc_num, attr_name);
	if (result < 0) {
		fprintf(stderr, "Failed to snprintf(path)\n");
		return 1;
	} else if (result > PATH_MAX) {
		fprintf(stderr, "Over PATH_MAX %d\n", mmc_num);
		return 2;
	}

	fp = fopen(path, "r");
	if (fp == NULL) {
		fprintf(stderr, "Failed to fopen() %s %s\n", path,
			strerror(errno));
		return 3;
	}

	res_fgets = fgets(attr, (int)size, fp);
	result = ferror(fp);

	if (fclose(fp) == EOF) {
		fprintf(stderr, "Failed to fclose() %d %s\n", mmc_num,
			strerror(errno));
		return 4;
	}

	if (res_fgets == NULL) {
		/* note: if significant error occurs such as setting errno, please retry this command. */
		if (result) {
			fprintf(stderr, "Could not read MMC attr file %s\n",
				path);
			return 5;
		} else {
			fprintf(stderr,
				"Could not read data from MMC attr file %s\n",
				path);
			return 6;
		}
	}

	attr[size - 1] = '\0';
	len = strnlen(attr, size);

	while (len > 0 && isspace(attr[len - 1]))
		len--;

	while (len > 0 && isspace(*start)) {
		start++;
		len--;
	}
	memmove(attr, start, len);
	attr[len] = '\0';

	return 0;
}

/**
 * Check device whether to support Micron specific commands.
 *  @param [in]  dev device file path. Set only eMMC 5.1 or later devices.
 *  @return Supported(true)/not Supported(false)
 */
static bool check_device(const char *dev)
{
	int result;
	int mmc_num;
	long mid, rev;
	char type[256];
	char attr[256];
	char *mmcblk_name, *dup_dev;

	dup_dev = strndup(dev, PATH_MAX);
	if (dup_dev == NULL) {
		fprintf(stderr, "Failed to strndup() %s\n", strerror(errno));
		return false;
	}

	mmcblk_name = basename(dup_dev);
	if (mmcblk_name == NULL) {
		fprintf(stderr, "Failed to basename()\n");
		goto endproc;
	}

	if (sscanf(mmcblk_name, "mmcblk%d", &mmc_num) != 1) {
		fprintf(stderr, "Failed to sscanf %s %s\n", dev,
			strerror(errno));
		goto endproc;
	}

	/** device type */
	result = read_attr(mmc_num, "type", type, 256);
	if (result)
		goto endproc;

	/** manufacturer ID */
	result = read_attr(mmc_num, "manfid", attr, 256);
	if (result)
		goto endproc;
	mid = strtol(attr, NULL, 16);
	if (mid == LONG_MAX || mid == LONG_MIN) {
		fprintf(stderr, "Failed to strtol %s %s\n", "manfid",
			strerror(errno));
		goto endproc;
	}

	/** eMMC revision */
	result = read_attr(mmc_num, "rev", attr, 256);
	if (result)
		goto endproc;
	rev = strtol(attr, NULL, 16);
	if (rev == LONG_MAX || rev == LONG_MIN) {
		fprintf(stderr, "Failed to strtol %s %s\n", "rev",
			strerror(errno));
		goto endproc;
	}

	if (use_verbose) {
		printf("device attrs:\n");
		printf("    device   mmc%d\n", mmc_num);
		printf("    type     %s\n", type);
		printf("    manf id  %lx\n", mid);
		printf("    revision %lx\n", rev);
	}

	free(dup_dev);
	return !strncmp(type, "MMC", 256) && mid == SUPPORTED_MID &&
	       rev >= SUPPORTED_REV;
endproc:
	free(dup_dev);
	return false;
}

static struct option options[] = { { "time", no_argument, 0, 't' },
				   { "help", no_argument, 0, 'h' },
				   { "verbose", no_argument, 0, 'v' },
				   { "version", no_argument, 0, 'V' },
				   { NULL, no_argument, NULL, 0 } };

static void usage(FILE *stream, int status)
{
	fprintf(stream, "Usage: emmc-sref [option(s)] [device file]\n");
	fprintf(stream,
		"  Initiate self refresh feature for Micron's eMMC device.\n");
	fprintf(stream, "  Options:\n");
	fprintf(stream, "  -t --time    Use system time resource.\n");
	fprintf(stream, "  -v --verbose Verbose message\n");
	fprintf(stream, "  -V --version Display version number of emmc-sref\n");
	fprintf(stream, "  -h --help    Display help message.\n");
	exit(status);
}

#if !defined(VERSION_STRING)
#define VERSION_STRING "for testing"
#endif

static void print_version(void)
{
	printf("emmc-sref %s\n", VERSION_STRING);
	printf("Copyright (c) 2021 Atmark Techno,Inc.\n");
}

int main(int argc, char **argv)
{
	int c;
	int result = 0;
	int use_system_time = 0;
	__u64 abs_time = 0;

	if (argc < 2) {
		usage(stderr, 1);
	}

	while ((c = getopt_long(argc, argv, "htvV", options, NULL)) != EOF) {
		switch (c) {
		case 0:
			break;
		case 'h':
			usage(stdout, 0);
			break;
		case 't':
			use_system_time = 1;
			break;
		case 'v':
			use_verbose = 1;
			break;
		case 'V':
			print_version();
			break;
		default:
			usage(stderr, 1);
			break;
		}
	}

	if (argv[optind] != NULL && !check_device(argv[optind])) {
		fprintf(stderr, "Device not supported.\n");
		result = 1;
	} else {
		if (use_system_time) {
			struct timeval now;
			result = gettimeofday(&now, NULL);
			if (result < 0) {
				fprintf(stderr, "Failed at getimeofday() %s\n",
					strerror(errno));
			} else {
				/* convert epoch time to Gregorian time */
				abs_time = (__u64)now.tv_sec + EPOCH_OFFSET;

				if (use_verbose)
					printf("system time date: %llu\n",
					       abs_time);
			}
		} else {
			if (argv[optind] != NULL) {
				result = get_time(argv[optind], &abs_time);
				if (abs_time == (__u64)UINT64_MAX)
					abs_time = 0;
				if (!result)
					abs_time += SECONDS_IN_A_DAY;
			}
		}

		if (abs_time && argv[optind] != NULL) {
			result = set_time(argv[optind], abs_time);
		}
	}

	return !result ? EXIT_SUCCESS : EXIT_FAILURE;
}
